
import ast

import pytest
from flake8_arg_spacing import ArgSpacingChecker


@pytest.mark.parametrize('arg, should_report', [
    ('x=- a', True), ('x=not a', True), ('x=await f()', True), ('x=a + b', True), ('x=a and b', True), ('x=a <= b', True), ('x=a if b else c', True),
    ('x=-a', False), ('x=a+b', False), ('x=a**b', False), ('x=a', False), ('x=(a + b)', False), ('x=f(a, b)', False), ('x=lambda: a', False),
    ('x = a + b', False), ('x: int', False), ('x: int = a + b', False)
])
@pytest.mark.parametrize('inner_template', [
    '{}',
    'y, {}',
    '*args, {}',
    '*args, y, {}',
    '{}, z=42',
    '{}, **kwargs',
    '{}, /',
])
@pytest.mark.parametrize('main_template, error_code', [
    ('f({})', 101),
    ('lambda {}: a + b', 102),
    ('async def f({}): pass', 102),
])
def test_checker(arg, should_report, inner_template, main_template, error_code):
    if error_code == 101:
        if 'x:' in arg:
            pytest.skip("Type annotation doesn't make sense for function calls")
        if ', /' in inner_template:
            pytest.skip("Positional-only syntax doesn't make sense for function calls")

    if main_template.startswith('lambda'):
        if 'x:' in arg:
            pytest.skip("Type annotation doesn't work in lambda definitions")

    source = main_template.format(inner_template.format(arg))
    extra_newlines = '\n' * (len(arg) - 3)
    tree = ast.parse(extra_newlines + source)
    errors = list(ArgSpacingChecker(tree).run())

    if should_report:
        assert len(errors) == 1
        error = errors[0]
        assert isinstance(error, tuple)
        assert len(error) == 4
        assert error[0] == len(extra_newlines) + 1
        assert error[1] == source.index(arg) + arg.index('=') + 1
        assert error[2] == f"ARG{error_code} confusing spacing around " + ("keyword argument" if error_code == 101 else "parameter default")
        assert error[3] is ArgSpacingChecker
    else:
        assert len(errors) == 0




import ast
import importlib.metadata
from itertools import chain


class ArgSpacingChecker:
    name = __name__
    version = importlib.metadata.version(__name__)

    def __init__(self, tree):
        self.tree = tree

    def _should_report(self, arg_node, value_node):
        # Don't report if there's a space after `=`.
        if arg_node.col_offset + len(arg_node.arg) + 1 < value_node.col_offset:
            return False

        # Report only if the given expression contains spaces.
        if isinstance(value_node, ast.UnaryOp):
            return value_node.col_offset + 2 <= value_node.operand.col_offset
        elif isinstance(value_node, ast.BinOp):
            return value_node.left.end_col_offset + 3 <= value_node.right.col_offset
        else:
            # Assume these are always written with spaces.
            return isinstance(value_node, (ast.Await, ast.BoolOp, ast.Compare, ast.IfExp))

    def _get_info(self, node, error_code, label):
        return (
            node.lineno,
            node.col_offset,
            f"ARG{error_code} confusing spacing around {label}",
            type(self),
        )

    def run(self):
        for node in ast.walk(self.tree):
            if isinstance(node, ast.keyword):
                if node.arg is not None and self._should_report(node, node.value):
                    yield self._get_info(node.value, 101, "keyword argument")

            if isinstance(node, ast.arguments):
                pos_args = node.posonlyargs + node.args
                pos_defaults = node.defaults
                pos_defaults_index = len(pos_args) - len(pos_defaults)
                kw_args = node.kwonlyargs
                kw_defaults = node.kw_defaults

                for arg, default in chain(zip(pos_args[pos_defaults_index:], pos_defaults), zip(kw_args, kw_defaults)):
                    if default is not None and self._should_report(arg, default):
                        yield self._get_info(default, 102, "parameter default")
